#include <stdint.h>

// Final unpack length: 168
// Decoded length: 42 (first four bytes of buffer)
const uint8_t muteButtonsPackedBMP[46] =
{
	0x00,0x00,0x00,0x2A,0xAF,0xAA,0xFE,0xAA,0xFA,0xAB,0xEA,0xBF,0xEA,0xAA,0xAF,0xFE,
	0xAA,0xBE,0xAF,0xEA,0xFA,0xAF,0xFF,0xAA,0xAA,0xFF,0xEA,0xAB,0xEB,0xFE,0xAA,0xBE,
	0xFF,0xEA,0xAA,0xAA,0xFE,0xAF,0xFA,0xFB,0xEB,0xFF,0xEA,0xBE,0xAA,0xAA
};
